"use strict";

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _typeof(obj) { return obj && typeof Symbol !== "undefined" && obj.constructor === Symbol ? "symbol" : typeof obj; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

var React = require("react");
var invariant = require("invariant");
var Uniform = require("../Uniform");
var Shaders = require("../Shaders");
var TextureObjects = require("./TextureObjects");
var isNonSamplerUniformValue = require("./isNonSamplerUniformValue");
var findGLNodeInGLComponentChildren = require("./findGLNodeInGLComponentChildren");
var unifyPropsWithContext = require("./unifyPropsWithContext");
var invariantStrictPositive = require("./invariantStrictPositive");

//// build: converts the gl-react VDOM DSL into an internal data tree.

module.exports = function build(GLNode, context, parentPreload, via, surfaceId, decorateOnShaderCompile) {
  var props = GLNode.props;
  var shader = Shaders._resolve(props.shader, surfaceId, decorateOnShaderCompile(props.onShaderCompile));
  var GLNodeUniforms = props.uniforms;

  var _unifyPropsWithContex = unifyPropsWithContext(props, context);

  var width = _unifyPropsWithContex.width;
  var height = _unifyPropsWithContex.height;
  var pixelRatio = _unifyPropsWithContex.pixelRatio;

  var newContext = {
    width: width,
    height: height,
    pixelRatio: pixelRatio
  };
  var GLNodeChildren = props.children;
  var preload = "preload" in props ? props.preload : parentPreload;

  invariant(Shaders.exists(shader), "Shader #%s does not exists", shader);

  var shaderName = Shaders.getName(shader);
  invariantStrictPositive(width, "GL Component (" + shaderName + "). width prop");
  invariantStrictPositive(height, "GL Component (" + shaderName + "). height prop");
  invariantStrictPositive(pixelRatio, "GL Component (" + shaderName + "). pixelRatio prop");

  var uniforms = _extends({}, GLNodeUniforms);
  var children = [];
  var contents = [];

  React.Children.forEach(GLNodeChildren, function (child) {
    invariant(child.type === Uniform, "(Shader '%s') GL.Node can only contains children of type GL.Uniform. Got '%s'", shaderName, child.type && child.type.displayName || child);
    var _child$props = child.props;
    var name = _child$props.name;
    var children = _child$props.children;

    var opts = _objectWithoutProperties(_child$props, ["name", "children"]);

    invariant(typeof name === "string" && name, "(Shader '%s') GL.Uniform must define an name String", shaderName);
    invariant(!GLNodeUniforms || !(name in GLNodeUniforms), "(Shader '%s') The uniform '%s' set by GL.Uniform must not be in {uniforms} props", shaderName);
    invariant(!(name in uniforms), "(Shader '%s') The uniform '%s' set by GL.Uniform must not be defined in another GL.Uniform", shaderName);
    uniforms[name] = !children || children.value ? children : { value: children, opts: opts }; // eslint-disable-line no-undef
  });

  Object.keys(uniforms).forEach(function (name) {
    var value = uniforms[name];
    var nonSamplerUniformTyp = isNonSamplerUniformValue(value);
    if (nonSamplerUniformTyp) {
      if (process.env.NODE_ENV !== "production" && nonSamplerUniformTyp === "number[]") {
        var i = value.length;
        while (i-- > 0 && !isNaN(value[i])) {}
        invariant(i < 0, "Shader '%s': uniform '%s' must be an array of numbers. Found '%s' at index %s", shaderName, name, value[i], i);
      }
      return;
    }

    var opts = undefined,
        typ = typeof value === "undefined" ? "undefined" : _typeof(value);

    if (value && typ === "object" && !value.prototype && "value" in value) {
      // if value has a value field, we tread this field as the value, but keep opts in memory if provided
      if (_typeof(value.opts) === "object") {
        opts = value.opts;
      }
      value = value.value;
      typ = typeof value === "undefined" ? "undefined" : _typeof(value);
    }

    if (!value) {
      // falsy value are accepted to indicate blank texture
      uniforms[name] = value;
    } else if (typ === "string") {
      // uri specified as a string
      uniforms[name] = TextureObjects.withOpts(TextureObjects.URI({ uri: value }), opts);
    } else if (typ === "object" && typeof value.uri === "string") {
      // uri specified in an object, we keep all other fields for RN "local" image use-case
      uniforms[name] = TextureObjects.withOpts(TextureObjects.URI(value), opts);
    } else if (typ === "object" && value.data && value.shape && value.stride) {
      // ndarray kind of texture
      uniforms[name] = TextureObjects.withOpts(TextureObjects.NDArray(value), opts);
    } else if (typ === "object" && (value instanceof Array ? React.isValidElement(value[0]) : React.isValidElement(value))) {
      // value is a VDOM or array of VDOM
      var res = findGLNodeInGLComponentChildren(value, newContext);
      if (res) {
        var childGLNode = res.childGLNode;
        var _via = res.via;
        // We have found a GL.Node children, we integrate it in the tree and recursively do the same

        children.push({
          vdom: value,
          uniform: name,
          data: build(childGLNode, newContext, preload, _via, surfaceId, decorateOnShaderCompile)
        });
      } else {
        // in other cases VDOM, we will use child as a content
        contents.push({
          vdom: value,
          uniform: name,
          opts: opts
        });
      }
    } else {
      // in any other case, it is an unrecognized invalid format
      delete uniforms[name];
      if (typeof console !== "undefined" && console.error) console.error("invalid uniform '" + name + "' value:", value); // eslint-disable-line no-console
      invariant(false, "Shader '%s': Unrecognized format for uniform '%s'", shaderName, name);
    }
  });

  return {
    shader: shader,
    uniforms: uniforms,
    width: width,
    height: height,
    pixelRatio: pixelRatio,
    children: children,
    contents: contents,
    preload: preload,
    via: via
  };
};
//# sourceMappingURL=build.js.map